"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var ts = __importStar(require("typescript"));
var type_1 = require("./type");
var flags_1 = require("./flags");
var config_1 = require("./config");
var processProperty = function (checker) { return function (s) {
    return s.name + ": " + processType(checker)(checker.getTypeOfSymbolAtLocation(s, s.valueDeclaration));
}; };
var getOptimizedStringLiteralUnion = function (type) {
    var unionTypes = type.types;
    return "t.keyof({" + unionTypes
        .map(function (t) { return "\"" + t.value + "\": null"; })
        .join(", ") + "})";
};
var processObjectType = function (checker) { return function (type) {
    var properties = checker.getPropertiesOfType(type);
    var requiredProperties = properties.filter(function (p) { return !p.valueDeclaration.questionToken; });
    var optionalProperties = properties.filter(function (p) { return p.valueDeclaration.questionToken; });
    if (requiredProperties.length && optionalProperties.length) {
        return "t.intersection([t.type({" + requiredProperties.map(processProperty(checker)) + "}), t.partial({" + optionalProperties
            .map(processProperty(checker))
            .join(", ") + "})])";
    }
    else if (optionalProperties.length === 0) {
        return "t.type({" + requiredProperties
            .map(processProperty(checker))
            .join(", ") + "})";
    }
    else {
        return "t.partial({" + optionalProperties
            .map(processProperty(checker))
            .join(", ") + "})";
    }
}; };
var processType = function (checker) { return function (type) {
    if (type_1.isLiteralType(type)) {
        return "t.literal(" + checker.typeToString(type) + ")";
    }
    else if (type_1.isPrimitiveType(type)) {
        return "t." + checker.typeToString(type);
    }
    else if (type_1.isBasicObjectType(type, checker)) {
        return "t.type({})";
    }
    else if (type_1.isRecordType(type)) {
        var _a = type.aliasTypeArguments, key = _a[0], value = _a[1];
        return "t.record(" + processType(checker)(key) + ", " + processType(checker)(value) + ")";
    }
    else if (type.isUnion()) {
        var isStringLiteralUnion = type.types.every(function (t) { return t.isStringLiteral(); });
        if (isStringLiteralUnion) {
            return getOptimizedStringLiteralUnion(type);
        }
        return "t.union([" + type.types.map(processType(checker)).join(", ") + "])";
    }
    else if (type.isIntersection()) {
        return "t.intersection([" + type.types
            .map(processType(checker))
            .join(", ") + "])";
    }
    else if (type_1.isTupleType(type, checker)) {
        if (type.hasRestElement) {
            console.warn("io-ts default validators do not support rest parameters in a tuple");
        }
        return "t.tuple([" + type.typeArguments.map(processType(checker)) + "])";
    }
    else if (type_1.isArrayType(type, checker)) {
        return "t.array(" + processType(checker)(type.getNumberIndexType()) + ")";
    }
    else if (type_1.isStringIndexedObjectType(type)) {
        return "t.record(t.string, " + processType(checker)(type.getStringIndexType()) + ")";
    }
    else if (type_1.isNumberIndexedType(type)) {
        return "t.record(t.number, " + processType(checker)(type.getNumberIndexType()) + ")";
    }
    else if (type_1.isFunctionType(type)) {
        return "t.Function";
    }
    else if (type_1.isObjectType(type)) {
        return processObjectType(checker)(type);
    }
    else if (type_1.isVoid(type)) {
        return "t.void";
    }
    else if (type_1.isAnyOrUnknown(type)) {
        return "t.unknown";
    }
    throw Error("Unknown type with type flags: " + flags_1.extractFlags(type.flags));
}; };
function handleDeclaration(node, checker) {
    var symbol, type;
    try {
        if (node.kind === ts.SyntaxKind.VariableStatement) {
            symbol = checker.getSymbolAtLocation(node.declarationList.declarations[0].name);
            type = checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration);
        }
        else {
            symbol = checker.getSymbolAtLocation(node.name);
            type = checker.getTypeAtLocation(node);
        }
        return "const " + symbol.name + " = " + processType(checker)(type);
    }
    catch (e) {
        return "// Error: Failed to generate a codec";
    }
}
var visit = function (checker, config, result) { return function (node) {
    if (!config.followImports &&
        !config.fileNames.includes(node.getSourceFile().fileName)) {
        return;
    }
    if (ts.isTypeAliasDeclaration(node) ||
        ts.isVariableStatement(node) ||
        ts.isInterfaceDeclaration(node)) {
        result.push(handleDeclaration(node, checker));
    }
    else if (ts.isModuleDeclaration(node)) {
        ts.forEachChild(node, visit(checker, config, result));
    }
}; };
var getImports = function () {
    return "import * as t from \"io-ts\"";
};
var compilerOptions = {
    strictNullChecks: true
};
function getValidatorsFromString(source, config) {
    if (config === void 0) { config = __assign(__assign({}, config_1.defaultConfig), { fileNames: [config_1.DEFAULT_FILE_NAME] }); }
    var defaultCompilerHostOptions = ts.createCompilerHost({});
    var compilerHostOptions = __assign(__assign({}, defaultCompilerHostOptions), { getSourceFile: function (filename, languageVersion) {
            var restArgs = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                restArgs[_i - 2] = arguments[_i];
            }
            if (filename === config_1.DEFAULT_FILE_NAME)
                return ts.createSourceFile(filename, source, ts.ScriptTarget.ES2015, true);
            else
                return defaultCompilerHostOptions.getSourceFile.apply(defaultCompilerHostOptions, __spreadArrays([filename,
                    languageVersion], restArgs));
        } });
    var program = ts.createProgram([config_1.DEFAULT_FILE_NAME], compilerOptions, compilerHostOptions);
    var checker = program.getTypeChecker();
    var result = config.includeHeader ? [getImports()] : [];
    ts.forEachChild(program.getSourceFile(config_1.DEFAULT_FILE_NAME), visit(checker, config, result));
    return result.join("\n\n");
}
exports.getValidatorsFromString = getValidatorsFromString;
function getValidatorsFromFileNames(config) {
    var program = ts.createProgram(config.fileNames, compilerOptions);
    var checker = program.getTypeChecker();
    var result = config.includeHeader ? [getImports()] : [];
    for (var _i = 0, _a = program.getSourceFiles(); _i < _a.length; _i++) {
        var sourceFile = _a[_i];
        if (!sourceFile.isDeclarationFile) {
            ts.forEachChild(sourceFile, visit(checker, config, result));
        }
    }
    return result.join("\n\n");
}
exports.getValidatorsFromFileNames = getValidatorsFromFileNames;
function isEntryPoint() {
    return require.main === module;
}
if (isEntryPoint()) {
    var config = config_1.getCliConfig();
    if (!config.fileNames.length) {
        config_1.displayHelp();
    }
    else {
        console.log(getValidatorsFromFileNames(config));
    }
}
//# sourceMappingURL=index.js.map